@@ -5,10 +5,27 @@ from __future__ import annotations
 import asyncio
 import json
 import time
-from typing import Any, Generic, Type, TypeAlias, TypeVar
+from typing import (
+    Any,
+    Coroutine,
+    Generic,
+    Protocol,
+    Type,
+    TypeAlias,
+    TypeVar,
+)
 
 import pandas as pd
 
+from qcog_python_client.log import qcoglogger
+from qcog_python_client.schema.common import Matrix, Model
+from qcog_python_client.schema.generated_schema.models import (
+    ModelEnsembleParameters,
+    ModelGeneralParameters,
+    ModelPauliParameters,
+    TrainingStatus,
+)
+
 from ._jsonable_parameters import (
     jsonable_inference_parameters,
     jsonable_train_parameters,
@@ -22,35 +39,47 @@ from .client import (
 from .schema import (
     AsyncInferenceProtocol,
     AsyncTrainProtocol,
-    Dataset,
-    EnsembleModel,
-    GeneralModel,
+    DatasetPayload,
     InferenceParameters,
     InferenceProtocol,
-    Model,
     NotRequiredStateParams,
     NotRequiredWeightParams,
     Operator,
-    PauliModel,
     TrainingParameters,
     TrainProtocol,
 )
 
-TrainingModel: TypeAlias = PauliModel | EnsembleModel | GeneralModel
-CLIENT = TypeVar("CLIENT")
+logger = qcoglogger.getChild(__name__)
 
 
-MODEL_MAP: dict[str, Type[TrainingModel]] = {
-    Model.pauli.value: PauliModel,
-    Model.ensemble.value: EnsembleModel,
-    Model.general.value: GeneralModel,
+TrainingModel: TypeAlias = (
+    ModelPauliParameters | ModelEnsembleParameters | ModelGeneralParameters
+)
+
+
+class MaybeAsyncClientProtocol(Protocol):
+    """A protocol in a disperate attemt to unify the async and sync clients."""
+
+    def get(self, *args: Any, **kwargs: Any) -> Coroutine | dict:
+        """Get method."""
+        ...
+
+
+CLIENT = TypeVar("CLIENT", bound=MaybeAsyncClientProtocol)
+
+
+ModelMap: dict[Model, Type[TrainingModel]] = {
+    Model.pauli: ModelPauliParameters,
+    Model.ensemble: ModelEnsembleParameters,
+    Model.general: ModelGeneralParameters,
 }
+
 # https://ubiops.com/docs/r_client_library/deployment_requests/#response-structure_1
-WAITING_STATUS = ("processing", "pending")
-SUCCESS_STATUS = "completed"
+WAITING_STATUS = (TrainingStatus.processing, TrainingStatus.pending)
+SUCCESS_STATUS = (TrainingStatus.completed,)
 
 
-DEFAULT_QCOG_VERSION = "0.0.63"
+DEFAULT_QCOG_VERSION = "0.0.89"
 
 
 def numeric_version(version: str) -> list[int]:
@@ -74,16 +103,98 @@ def numeric_version(version: str) -> list[int]:
     return [int(w) for w in numbers]
 
 
+class StateNotSetError(Exception):
+    """Exception raised when a state is not set."""
+
+    def __init__(self, missing_state: str, suggestion: str | None = None) -> None:
+        """Initialize the exception.
+
+        Attributes
+        ----------
+        missing_state : str
+            The state that is missing.
+
+        suggestion : str | None
+            A suggestion for remediation.
+
+        """
+        self.missing_state = missing_state
+        self.suggestion = suggestion
+        super().__init__(f"Missing state: {missing_state}. {suggestion or ''}")
+
+
 class BaseQcogClient(Generic[CLIENT]):  # noqa: D101
     def __init__(self) -> None:  # noqa: D107
         self._version: str
         self._http_client: CLIENT
-        self.model: TrainingModel
-        self.project: dict[str, str]
-        self.dataset: dict = {}
-        self.training_parameters: dict = {}
-        self.trained_model: dict = {}
-        self.inference_result: dict = {}
+        self._model: TrainingModel | None = None
+        self._project: dict | None = None
+        self._dataset: dict | None = None
+        self._training_parameters: dict | None = None
+        self._trained_model: dict | None = None
+        self._inference_result: dict | None = None
+        self._loss: Matrix | None = None
+
+    @property
+    def model(self) -> TrainingModel:
+        """Return the model."""
+        if self._model is None:
+            raise StateNotSetError(
+                "model", "Please set the model first using `pauli` or `ensemble` method"
+            )
+        return self._model
+
+    @property
+    def project(self) -> dict:
+        """Return the project."""
+        if self._project is None:
+            raise StateNotSetError(
+                "No project has been found associated with this request."
+            )
+        return self._project
+
+    @property
+    def dataset(self) -> dict:
+        """Return the dataset."""
+        if self._dataset is None:
+            raise StateNotSetError(
+                "No dataset has been found associated with this request.",
+                """You can use `data` method to upload a new dataset or
+                `preloaded_data` to load an existing one""",
+            )
+        return self._dataset
+
+    @property
+    def training_parameters(self) -> dict:
+        """Return the training parameters."""
+        if self._training_parameters is None:
+            raise StateNotSetError(
+                "No training parameters have been found associated with this request.",
+                """You can use `train` method to start a new training or
+                `preloaded_training_parameters` to load existing ones""",  # noqa: 501
+            )
+        return self._training_parameters
+
+    @property
+    def trained_model(self) -> dict:
+        """Return the trained model."""
+        if self._trained_model is None:
+            raise StateNotSetError(
+                "No trained model has been found associated with this request.",
+                """You can use `train` method to start a new training or
+                `preloaded_model` to load an existing one""",
+            )
+        return self._trained_model
+
+    @property
+    def inference_result(self) -> dict:
+        """Return the inference result."""
+        if self._inference_result is None:
+            raise StateNotSetError(
+                "No inference result has been found associated with this request.",
+                "You can use `inference` method to run an inference",
+            )
+        return self._inference_result
 
     @property
     def version(self) -> str:
@@ -115,14 +226,15 @@ class BaseQcogClient(Generic[CLIENT]):  # noqa: D101
         target_operator: list[Operator] = [],
     ) -> Any:
         """Select PauliModel for the training."""
-        self.model = PauliModel(
-            operators,
-            qbits,
-            pauli_weight,
-            sigma_sq,
-            sigma_sq_optimization,
-            seed,
-            target_operator,
+        self._model = ModelPauliParameters(
+            operators=[str(op) for op in operators],
+            qbits=qbits,
+            pauli_weight=pauli_weight,
+            sigma_sq=sigma_sq,
+            sigma_sq_optimization_kwargs=sigma_sq_optimization,
+            seed=seed,
+            target_operators=[str(op) for op in target_operator],
+            model_name=Model.pauli.value,
         )
         return self
 
@@ -137,17 +249,104 @@ class BaseQcogClient(Generic[CLIENT]):  # noqa: D101
         target_operator: list[Operator] = [],
     ) -> Any:
         """Select EnsembleModel for the training."""
-        self.model = EnsembleModel(
-            operators,
-            dim,
-            num_axes,
-            sigma_sq,
-            sigma_sq_optimization,
-            seed,
-            target_operator,
+        # Cast all the operators to string
+        self._model = ModelEnsembleParameters(
+            operators=[str(op) for op in operators],
+            dim=dim,
+            num_axes=num_axes,
+            sigma_sq=sigma_sq,
+            sigma_sq_optimization_kwargs=sigma_sq_optimization,
+            seed=seed,
+            target_operators=[str(op) for op in target_operator],
+            model_name=Model.ensemble.value,
         )
         return self
 
+    async def _progress(self) -> dict:
+        """Return the current status of the training.
+
+        Returns
+        -------
+        dict
+            the current status of the training
+            `guid` : str
+            `training_completion` : int
+            `current_batch_completion` : int
+            `status` : TrainingStatus
+
+        """
+        maybe_awaitable = self.http_client.get(f"model/{self.trained_model['guid']}")
+
+        if asyncio.iscoroutine(maybe_awaitable):
+            self._trained_model = await maybe_awaitable
+        else:
+            # Local VSCode MyPy plugin, correctly recognizes that
+            # maybe_awaitable is not a coroutine, but, when
+            # running the command, it throws an error.
+            self._trained_model = maybe_awaitable
+
+        return {
+            "guid": self.trained_model.get("guid"),
+            "training_completion": self.status_resp.get("training_completion"),
+            "current_batch_completion": self.trained_model.get(
+                "current_batch_completion"
+            ),
+            "status": TrainingStatus(self.trained_model.get("status")),
+        }
+
+    async def _status(self) -> TrainingStatus:
+        """Fetch the status of the training request."""
+        maybe_awaitable = self.http_client.get(f"model/{self.trained_model['guid']}")
+
+        if asyncio.iscoroutine(maybe_awaitable):
+            self.status_resp = await maybe_awaitable
+        else:
+            self.status_resp = maybe_awaitable
+
+        status = TrainingStatus(self.status_resp["status"])
+        training_completion = self.status_resp.get("training_completion", 0)
+        current_batch_completion = self.status_resp.get("current_batch_completion", 0)
+
+        logger.info(f"\nStatus: {status}")
+        logger.info(f"Training completion: {training_completion} %")
+        logger.info(f"Current batch completion: {current_batch_completion} %\n")
+
+        self.last_status = TrainingStatus(self.status_resp["status"])
+
+        # Try to set the _loss attribute if present
+        self._loss = self.status_resp.get("loss")
+
+        return self.last_status
+
+    async def _get_loss(self) -> Matrix | None:
+        """Return the loss for a trained model.
+
+        If the training is not completed, the loss will not be available.
+        """
+        # Fetch and set a loss if it's not already set
+        if self._loss is None:
+            maybe_awaitable = self.http_client.get(
+                f"model/{self.trained_model['guid']}"
+            )
+
+            # maybe_awaitable: in order to avoid code duplication
+            # on the async and the sync client, we will check if the
+            # method has been called from the async client or not.
+
+            # If the method has been called from the async client,
+            # the `http_client.get` will return a coroutine,
+            if asyncio.iscoroutine(maybe_awaitable):
+                self.status_resp = await maybe_awaitable
+                # Otherwise it will return a dict
+            else:
+                self.status_resp = maybe_awaitable
+
+            # Set the last status
+            self.last_status = TrainingStatus(self.status_resp["status"])
+            # Try to set the _loss attribute if present
+            self._loss = self.status_resp.get("loss")
+        return self._loss
+
 
 class QcogClient(  # noqa: D101
     BaseQcogClient[RequestsClient],
@@ -266,13 +465,12 @@ class QcogClient(  # noqa: D101
             hostname=hostname,
             port=port,
             api_version=api_version,
+            qcog_version=version,
         )
 
         if safe_mode:
             hsm.http_client.get("status")
 
-        hsm.project = hsm.http_client.get("bootstrap")
-
         return hsm
 
     def _preload(self, ep: str, guid: str) -> dict:
@@ -293,7 +491,7 @@ class QcogClient(  # noqa: D101
         """
         return self.http_client.get(f"{ep}/{guid}")
 
-    def _training_parameters(self, params: TrainingParameters) -> None:
+    def _preload_training_parameters(self, params: TrainingParameters) -> None:
         """Upload training parameters.
 
         Parameters
@@ -302,12 +500,11 @@ class QcogClient(  # noqa: D101
             Valid TypedDict of the training parameters
 
         """
-        self.training_parameters = self.http_client.post(
+        self._training_parameters = self.http_client.post(
             "training_parameters",
             {
-                "project_guid": self.project["guid"],
-                "model": self.model.value,
-                "parameters": {"model": self.model.params}
+                "model": self.model.model_name,
+                "parameters": {"model": self.model.model_dump()}
                 | jsonable_train_parameters(params),
             },
         )
@@ -330,14 +527,16 @@ class QcogClient(  # noqa: D101
         AsyncQcogClient
 
         """
-        data_payload = Dataset(
+        # Validate data payload
+        data_payload = DatasetPayload(
             format="dataframe",
             source="client",
             data=encode_base64(data),
-            project_guid=self.project["guid"],
-        )
+            project_guid=None,
+        ).model_dump()
+
         valid_data: dict = {k: v for k, v in data_payload.items()}  # type cast
-        self.dataset = self.http_client.post("dataset", valid_data)
+        self._dataset = self.http_client.post("dataset", valid_data)
         return self
 
     def preloaded_data(self, guid: str) -> QcogClient:
@@ -354,7 +553,7 @@ class QcogClient(  # noqa: D101
             itself
 
         """
-        self.dataset = self._preload("dataset", guid)
+        self._dataset = self._preload("dataset", guid)
         return self
 
     def preloaded_training_parameters(
@@ -376,7 +575,7 @@ class QcogClient(  # noqa: D101
             itself
 
         """
-        self.training_parameters = self._preload(
+        self._training_parameters = self._preload(
             "training_parameters",
             guid,
         )
@@ -384,22 +583,23 @@ class QcogClient(  # noqa: D101
 
     def preloaded_model(self, guid: str) -> QcogClient:
         """Preload a model from a guid."""
-        self.trained_model = self._preload("model", guid)
+        self._trained_model = self._preload("model", guid)
 
-        self._preload("project", self.trained_model["project_guid"])
         self.version = self.trained_model["qcog_version"]
 
         self.preloaded_training_parameters(
             self.trained_model["training_parameters_guid"]
         )
 
-        model_params = {
-            k.replace("_kwargs", ""): v
-            for k, v in self.training_parameters["parameters"]["model"].items()
-            if k != "model"
-        }
+        # The model name in order to retrieve the correct parameters
+        model_name = Model(self.training_parameters["model"])
 
-        self.model = MODEL_MAP[self.training_parameters["model"]](**model_params)
+        # Parameters for the model
+        model_parameters = self.training_parameters["parameters"]["model"]
+        # Retrieve the validation for the model parameters
+        validate_cls = ModelMap[model_name]
+        # Validate the model parameters
+        self._model = validate_cls(**model_parameters)
         return self
 
     def train(
@@ -437,28 +637,37 @@ class QcogClient(  # noqa: D101
             state_kwargs=get_states_extra,
         )
 
-        self._training_parameters(params)
+        self._preload_training_parameters(params)
 
-        self.trained_model = self.http_client.post(
+        self._trained_model = self.http_client.post(
             "model",
             {
                 "training_parameters_guid": self.training_parameters["guid"],
                 "dataset_guid": self.dataset["guid"],
-                "project_guid": self.project["guid"],
                 "qcog_version": self.version,
             },
         )
 
         return self
 
-    def status(self) -> str:
+    def status(self) -> TrainingStatus:
         """Fetch the status of the training request."""
-        self.status_resp: dict = self.http_client.get(
-            f"model/{self.trained_model['guid']}"
-        )
+        return asyncio.run(self._status())
 
-        self.last_status: str = self.status_resp["status"]
-        return self.last_status
+    def progress(self) -> dict:
+        """Return the current status of the training.
+
+        Returns
+        -------
+        dict
+            the current status of the training
+            `guid` : str
+            `training_completion` : int
+            `current_batch_completion` : int
+            `status` : TrainingStatus
+
+        """
+        return asyncio.run(self._progress())
 
     def wait_for_training(self, poll_time: int = 60) -> QcogClient:
         """Wait for training to complete.
@@ -478,7 +687,10 @@ class QcogClient(  # noqa: D101
             itself
 
         """
-        while self.status() in WAITING_STATUS:
+        logger.info("Waiting for training to complete")
+
+        while s := self.status() in WAITING_STATUS:
+            logger.info(f"Current status: {s}")
             if self.last_status in SUCCESS_STATUS:
                 break
             time.sleep(poll_time)
@@ -519,8 +731,6 @@ class QcogClient(  # noqa: D101
             },
         )
 
-        print(inference_result)
-
         return base642dataframe(
             inference_result["response"]["data"],
         )
@@ -569,6 +779,13 @@ class QcogClient(  # noqa: D101
 
         return self
 
+    def loss(self) -> Matrix | None:
+        """Return the loss for a trained model.
+
+        If the training is not completed, the loss will not be available.
+        """
+        return asyncio.run(self._get_loss())
+
 
 class AsyncQcogClient(  # noqa: D101
     BaseQcogClient[AIOHTTPClient],
@@ -648,18 +865,12 @@ class AsyncQcogClient(  # noqa: D101
             hostname=hostname,
             port=port,
             api_version=api_version,
+            qcog_version=version,
         )
 
         if safe_mode:
             await hsm.http_client.get("status")
 
-        hsm.project = RequestsClient(
-            token=token,
-            hostname=hostname,
-            port=port,
-            api_version=api_version,
-        ).get("bootstrap")
-
         return hsm
 
     async def _preload(self, ep: str, guid: str) -> dict:
@@ -680,7 +891,7 @@ class AsyncQcogClient(  # noqa: D101
         """
         return await self.http_client.get(f"{ep}/{guid}")
 
-    async def _training_parameters(self, params: TrainingParameters) -> None:
+    async def _preload_training_parameters(self, params: TrainingParameters) -> None:
         """Upload training parameters.
 
         Parameters
@@ -689,12 +900,11 @@ class AsyncQcogClient(  # noqa: D101
             Valid TypedDict of the training parameters
 
         """
-        self.training_parameters = await self.http_client.post(
+        self._training_parameters = await self.http_client.post(
             "training_parameters",
             {
-                "project_guid": self.project["guid"],
-                "model": self.model.value,
-                "parameters": {"model": self.model.params}
+                "model": self.model.model_name,
+                "parameters": {"model": self.model.model_dump()}
                 | jsonable_train_parameters(params),
             },
         )
@@ -718,14 +928,16 @@ class AsyncQcogClient(  # noqa: D101
             itself
 
         """
-        data_payload = Dataset(
+        # Validate data payload
+        data_payload = DatasetPayload(
             format="dataframe",
             source="client",
             data=encode_base64(data),
-            project_guid=self.project["guid"],
-        )
-        valid_data: dict = {k: v for k, v in data_payload.items()}  # type cast
-        self.dataset = await self.http_client.post("dataset", valid_data)
+            project_guid=None,
+        ).model_dump()
+
+        valid_data: dict = {k: v for k, v in data_payload.items()}
+        self._dataset = await self.http_client.post("dataset", valid_data)
         return self
 
     async def preloaded_data(self, guid: str) -> AsyncQcogClient:
@@ -742,7 +954,7 @@ class AsyncQcogClient(  # noqa: D101
             itself
 
         """
-        self.dataset = await self._preload("dataset", guid)
+        self._dataset = await self._preload("dataset", guid)
         return self
 
     async def preloaded_training_parameters(
@@ -764,7 +976,7 @@ class AsyncQcogClient(  # noqa: D101
             itself
 
         """
-        self.training_parameters = await self._preload(
+        self._training_parameters = await self._preload(
             "training_parameters",
             guid,
         )
@@ -772,24 +984,23 @@ class AsyncQcogClient(  # noqa: D101
 
     async def preloaded_model(self, guid: str) -> AsyncQcogClient:
         """Preload a model from a guid."""
-        self.trained_model = await self._preload("model", guid)
+        self._trained_model = await self._preload("model", guid)
 
         self.version = self.trained_model["qcog_version"]
 
-        await asyncio.gather(
-            self._preload("project", self.trained_model["project_guid"]),
-            self.preloaded_training_parameters(
-                self.trained_model["training_parameters_guid"]
-            ),
+        await self.preloaded_training_parameters(
+            self.trained_model["training_parameters_guid"]
         )
 
-        model_params = {
-            k.replace("_kwargs", ""): v
-            for k, v in self.training_parameters["parameters"]["model"].items()
-            if k != "model"
-        }
+        # The model name in order to retrieve the correct parameters
+        model_name = Model(self.training_parameters["model"])
 
-        self.model = MODEL_MAP[self.training_parameters["model"]](**model_params)
+        # Parameters for the model
+        model_parameters = self.training_parameters["parameters"]["model"]
+        # Retrieve the validation for the model parameters
+        validate_cls = ModelMap[model_name]
+        # Validate the model parameters
+        self._model = validate_cls(**model_parameters)
         return self
 
     async def train(
@@ -827,28 +1038,37 @@ class AsyncQcogClient(  # noqa: D101
             state_kwargs=get_states_extra,
         )
 
-        await self._training_parameters(params)
+        await self._preload_training_parameters(params)
 
-        self.trained_model = await self.http_client.post(
+        self._trained_model = await self.http_client.post(
             "model",
             {
                 "training_parameters_guid": self.training_parameters["guid"],
                 "dataset_guid": self.dataset["guid"],
-                "project_guid": self.project["guid"],
                 "qcog_version": self.version,
             },
         )
 
         return self
 
-    async def status(self) -> str:
-        """Fetch the status of the training request."""
-        self.status_resp: dict = await self.http_client.get(
-            f"model/{self.trained_model['guid']}"
-        )
+    async def status(self) -> TrainingStatus:
+        """Fetch the status of the training request. wrapped in a coroutine."""
+        return await self._status()
 
-        self.last_status: str = self.status_resp["status"]
-        return self.last_status
+    async def progress(self) -> dict:
+        """Return the current status of the training.
+
+        Returns
+        -------
+        dict
+            the current status of the training
+            `guid` : str
+            `training_completion` : int
+            `current_batch_completion` : int
+            `status` : TrainingStatus
+
+        """
+        return await self._progress()
 
     async def wait_for_training(self, poll_time: int = 60) -> AsyncQcogClient:
         """Wait for training to complete.
@@ -864,7 +1084,11 @@ class AsyncQcogClient(  # noqa: D101
             itself
 
         """
+        logger.info("Waiting for training to complete")
+
         while (await self.status()) in WAITING_STATUS:
+            logger.info(f"Current status: {self.last_status}")
+
             if self.last_status in SUCCESS_STATUS:
                 break
             await asyncio.sleep(poll_time)
@@ -897,7 +1121,7 @@ class AsyncQcogClient(  # noqa: D101
             the predictions
 
         """
-        self.inference_result: dict = await self.http_client.post(
+        self._inference_result: dict = await self.http_client.post(
             f"model/{self.trained_model['guid']}/inference",
             {
                 "data": encode_base64(data),
@@ -909,8 +1133,15 @@ class AsyncQcogClient(  # noqa: D101
             self.inference_result["response"]["data"],
         )
 
+    async def loss(self) -> Matrix | None:
+        """Return the loss for a trained model.
+
+        If the training is not completed, the loss will not be available.
+        """
+        return await self._get_loss()
+
     def pauli(  # noqa: D102
-self,
+        self,
         operators: list[Operator],
         qbits: int = 2,
         pauli_weight: int = 2,
