@@ -123,11 +123,12 @@ def test_train_one():
         get_states_extra=get_states_extra,
     )
 
+    print("*** WAITING FOR TRAINING ***")
     qcml.wait_for_training(poll_time=10)
     print("----- MODEL TRAINED -----")
     print(qcml.trained_model["guid"])
     print("----- LOSS -----")
-    print(qcml.loss)
+    print(qcml.loss())
 
 
 def test_inference_one():
